ll(core, "Everything that isn't part of another subsystem")
ll(spandsp, "Log messages generated by SpanDSP directly")
ll(ffmpeg, "Log messages generated by ffmpeg directly")
ll(transcoding, "Media and RTP transcoding")
ll(codec, "Codec negotiation")
ll(rtcp, "RTCP handling")
ll(ice, "ICE negotiation")
ll(crypto, "Negotiation of SRTP, crypto suites, DTLS, SDES")
ll(srtp, "SRTP encryption and decryption")
ll(internals, "Noisy low-level internals")
ll(http, "HTTP, HTTPS, Websockets")
ll(control, "Control protocols including SDP exchanges, CLI")
ll(dtx, "DTX timer/buffer")
